# HOWTO using SPICE in QEMU with encryption

## Introduction

By default, SPICE traffic is not encrypted. Obviously, this is a major concern in term of security.
This document describes how to generate and use TLS certificates in order to provide encryption over SPICE traffic.

## gen-cert.sh

The `gen-cert.sh` script, located in the `config/certs` directory, is used to generate TLS certificates for the Nexus project.
It is inspired by this page from the QEMU project: https://qemu-project.gitlab.io/qemu/system/tls.html
The script requires `certtool` which is part of the `gnutls-bin` Debian package. It can be installed with:
```
sudo apt-get install gnutls-bin
```

### Templates

`gen-cert.sh` uses template files to describe the service. Two templates are required:

 - `ca.info`: describes the Certification Authority (CA).
 - `nexus-server.info`: describes the SPICE server; in case of multiple servers, there must be one `.info` per server.

Note: several identity servers can be used and QEMU must be configured to point to one with the `x509-dir` argument:
```
qemu-system-x86_64 -spice tls-port=5930,x509-dir=./server-nexus-01
```

### Templates format

Format of the CA template file (`ca.info`):
```
cn = My Awesome Organisation
ca
cert_signing_key
```

<!--
Historically, the `cn` field must be a unique name. This field should not be used anymore.
It will be replaced by `dns_name` and `ip_address` which can be used multiple times.
-->

Format of the server template file (`nexus-server.info`):
```
organization = My Awesome Organisation
cn = my-nexus-server
dns_name = my-nexus-server
ip_address = 192.168.1.10
tls_www_server
encryption_key
signing_key
```

Remarks:

- `ip_address` could be an ipv6 address
- multiple `dns_name` lines can be specified

To obtain the fully qualified domain name (FQDN) of a system:
```
hostname --fqdn
```

### Methodology

Using both template files, `gen-cert.sh` creates in the specified directory:

- The CA master certificate: `ca-cert.pem`
- The server certificate signed with ca-cert.pem: `server-cert.pem`
- The server private key: `server-key.pem` (root-only access)

This directory must be passed in argument to QEMU in order to use SPICE with TLS.
The `ca-key.pem` key is sensitive and required to generate other certificates. Ideally, it whould be stored offline.
Each client must use the public key `ca-cert.pem`.

## Client-side certificate usage

On the client side, the public key (`ca-cert.pem`) can be used in two different ways:

(1) Passed as an argument to the SPICE client
(1) Installed, system-wide, on the client system (requires root access)

### (1) Pass the public key to the SPICE client

Using `remote-viewer` without password:
```
remote-viewer spice://server_hostname?tls-port=5930 --spice-ca-file=./ca-cert.pem --spice-secure-channels=all
```

Using `spicy` without password:
```
spicy -h server_hostname -s 5930 --spice-ca-file=ca-cert.pem --spice-secure-channels=all
```
Using `spicy`, with a password:
```
spicy -h server_hostname -s 5930 --spice-ca-file=ca-cert.pem -w passwd --spice-secure-channels=all
```

### (2) Install the public key system-wide

On Debian-like distributions, the public key can be installed using the `update-ca-certificates` tool.
More information on how to do it here: `https://ubuntu.com/server/docs/security-trust-store`

## Server side certificate usage

On the server side, SPICE can be configured to either use a password or not when establishing the connection.
In both cases, a directory containing these 3 files, `ca-cert.pem, server-cert.pem, server-key.pem`, must be specified to QEMU.

### SPICE without password

To start QEMU with SPICE+TLS, without password:
```
qemu-system-x86_64 -spice tls-port=5930,disable-ticketing=on,x509-dir=./nexus-server
```

The specified directory `./nexus-server` contains the files described above.

### SPICE with password

There are two ways to specify a password for SPICE:

- In clear, on the QEMU command line
- As a QEMU secret stored in a file

Here is how to specify the password (`12345678`) on the QEMU command line:
```
qemu-system-x86_64 -spice tls-port=5930,password=12345678,x509-dir=./server-nexus
```

Specifying a password as a QEMU secret is a bit more involving. The procedure is described in the next section.

### SPICE with password as QEMU secret

A QEMU secret defines an object associated to an id.
There are three ways to specify a secret with QEMU (here `mysecret0` with password `12345678`):

1. In clear:
   ```
   -object secret,id=mysecret0,data=12345678
   ```
1. In clear, encoded in base64:
   ```
   -object secret,id=mysecret0,data=MTIzNDU2Nzg=,format=base64
   ```
1. In a file (here `secret.txt`), in clear (important: the secret is the whole file's content, including any potential `\n`):
   ```
   echo -n 12345678 > secret.txt
   -object secret,id=mysecret0,file=secret.txt
   ```
1. In a file (here `secret.txt`), in base64:
   ```
   echo -n 12345678 | base64 > secret.txt
   -object secret,id=mysecret0,file=secret.txt,format=base64
   ```

For instance, the final QEMU command line for the secret `mysecret0` whose password is stored in Base64 in `secret.txt` would be:
```
qemu-system-x86_64 -object secret,id=mysecret0,file=secret.txt,format=base64 -spice tls-port=5930,password-secret=mysecret0,x509-dir=./server-nexus
```

<!--
## List of files on the server

```
├── backing
│   ├── ubuntu_srv-20.04-disk000.qcow2
├── images
│   ├── vm-test1_spice.qcow2
│   ├── vm-test2_spice.qcow2
│   └── vm-test3_spice.qcow2
└── pki
    ├── README.md
    ├── ca-cert.pem
    ├── ca-key.pem
    ├── ca.info
    ├── gen-cert.sh
    ├── passwd_vm1.txt
    ├── passwd_vm2.txt
    ├── passwd_vm3.txt
    ├── server-nexus-01.info
    ├── server-nexus-02.info
    ├── server-nexus-01
    │   ├── ca-cert.pem
    │   ├── server-cert.pem
    │   └── server-key.pem
    └── server-nexus-02
        ├── ca-cert.pem
        ├── server-cert.pem
        └── server-key.pem
```
-->
