package vms

import(
    "os"
    "sort"
    "path"
    "sync"
    "errors"
    "path/filepath"
    t "nexus-common/template"
    "nexus-server/utils"
    "github.com/google/uuid"
)

type (
    TemplateKeeperFn func(tpl *Template) bool

    Templates struct {
        m map[string]*Template
        dir string            // Base directory where templates reside
        rwlock *sync.RWMutex
    }
)

var templates *Templates

// Returns a Templates singleton to access this module's public functions.
// IMPORTANT: the InitTemplates function must have been previously called!
// Concurrency: safe
func GetTemplatesInstance() *Templates {
    return templates
}

// Create all templates from on-disk files.
// REMARK: path is the root directory where templates reside.
// Concurrency: unsafe
func InitTemplates() error {
    templatesDir := conf.TemplatesDir
    templates = &Templates { m: make(map[string]*Template), dir: templatesDir, rwlock: new(sync.RWMutex) }

    dirs, err := utils.GetSubDirs(templatesDir)
    if err != nil {
        return errors.New("Failed reading templates directory: "+err.Error())
    }

    for _, dir := range(dirs) {
        // Checks the template directory is valid.
        templateDir := path.Base(dir)
        _, err := uuid.Parse(templateDir)
        if err != nil {
            log.Warn("Skipping template "+templateDir+": invalid directory name")
            continue
        }

        // Checks that a templateDiskFile is present
        if _, err := os.Stat(filepath.Join(dir, templateDiskFile)); err != nil {
            log.Warn("Skipping template "+templateDir+": missing "+templateDiskFile)
            continue
        }

        // Read and validate the template's content
        template, err := newTemplateFromFile(filepath.Join(dir, templateConfFile))
        if err != nil {
            log.Warn("Skipping template "+templateDir+": "+err.Error())
            continue
        }

        templates.m[templateDir] = template
    }

    return nil
}

// Returns a list of templates for which TemplateKeeperFn is true.
// Concurrency: safe
func (templates *Templates)GetNetworkSerializedTemplates(keepFn TemplateKeeperFn) []t.TemplateSerialized {
    templates.rwlock.RLock()

    list := []t.TemplateSerialized{}
    for _, tpl := range templates.m {
        if keepFn(tpl) {
            list = append(list, tpl.SerializeToNetwork())
        }
    }
    templates.rwlock.RUnlock()

    // Sort templates by names
    sort.Slice(list, func(i, j int) bool {
        return list[i].Name < list[j].Name
    })
    return list
}

// Returns a template by its ID. Concurrency-safe version.
// Concurrency: safe
func (templates *Templates)GetTemplate(tplID uuid.UUID) (*Template, error) {
    templates.rwlock.RLock()
    defer templates.rwlock.RUnlock()
    return templates.getTemplateUnsafe(tplID)
}

// Returns a template by its ID. Concurrency-unsafe version.
// Concurrency: unsafe
func (templates *Templates)getTemplateUnsafe(tplID uuid.UUID) (*Template, error) {
    template, exists := templates.m[tplID.String()]
    if !exists {
        return nil, errors.New("Template not found")
    }
    return template, nil
}

// Deletes a template by its ID.
// A template can only be deleted if no VM references it!
// TODO: prevents deletion if the template's disk image is
//       being exported by RouterTemplates.ExportDisk
// Concurrency: safe
func (templates *Templates)DeleteTemplate(tplID uuid.UUID, vms *VMs) error {
    id := tplID.String()

    templates.rwlock.Lock()
    defer templates.rwlock.Unlock()
    template, exists := templates.m[id]
    if exists {
        if vms.IsTemplateUsed(id) {
            return errors.New("Failed: template still used by one or more VMs")
        }
        // Removes the template's files (and directory).
        if err := template.delete(); err != nil {
            return err
        }
        // Removes the template from the map.
        delete(templates.m, id)
        return nil
    }

    return errors.New("Failed: template not found")
}

// Adds a template, create its on-disk directory and config file.
// Concurrency: safe
func (templates *Templates)AddTemplate(template *Template) error {
    templates.rwlock.Lock()
    defer templates.rwlock.Unlock()

    err := template.writeConfig()
    if err != nil {
        return err
    }

    // Adds template to the map of templates.
    templates.m[template.t.ID.String()] = template

    return nil
}

// Edit a template's fields (name and/or access) and updates its config file accordingly.
// Concurrency: safe
func (templates *Templates)EditTemplate(tplID uuid.UUID, name, access string) error {
    templates.rwlock.RLock()
    defer templates.rwlock.RUnlock()

    tpl, err := templates.getTemplateUnsafe(tplID)
    if err != nil {
        return err
    }

    // Only updates fields that have changed.
    oriVal := tpl.t  // Saves original template values.

    if name != "" {
        tpl.t.Name = name
    }
    if access != "" {
        tpl.t.Access = access
    }

    if err = tpl.validate(); err != nil {
        // Restores original template values.
        tpl.t = oriVal
        return err
    }

    err = tpl.writeConfig()
    if err != nil {
        return err
    }

    return nil
}
